from .models import Project, MypyRunLineItem
from rich.table import Table


def format_projects(projects: list[Project]) -> Table:
    table = Table(title=f"Projects")
    if projects:
        table.add_column("ID")
        table.add_column("Name")
        table.add_column("Description")
        table.add_column("URL")

        for project in projects:
            table.add_row(
                f"{project.id}",
                f"{project.name}",
                f"{project.description}",
                f"{project.url}",
            )
    else:
        table.add_column("No Projects")

    return table


def format_mypy_items(run_id: int, items: list[MypyRunLineItem]) -> Table:
    table = Table(title=f"Mypy Coverage for Run ID: {run_id}")
    table.add_column("File")
    table.add_column("Imprecision")
    table.add_column("Lines")

    total_imprecision, total_loc = 0, 0
    for item in items:
        table.add_row(item.path, f"{item.imprecision} %", f"{item.loc}")
        total_imprecision += item.imprecision
        total_loc += item.loc
    avg = total_imprecision / len(items)
    table.add_row("--", "--", "--")
    table.add_row("Summary", f"{avg:.2f} %", f"{total_loc}")
    table.add_row("--", "--", "--")
    return table



"""CLI
"""
import click
import sqlalchemy
from pathlib import Path
import os
import sys
from code_census.models import (
    Project,
    Run,
    MypyRunLineItem,
    create_session,
    create_engine,
    get_project,
    add_run,
    add_mypy_line_items,
    get_mypy_line_items_by_run_id,
    get_runs,
    get_projects,
)
from code_census.mypy_utils import get_type_coverage
from code_census import formatter
from alembic.config import Config
from alembic import command
from rich.console import Console
from rich.table import Table
import os
import json


HELP_TEXT = {
    "db_url": (
        "Pass postgres connection details like postgresql://postgres:pass@db.host:5432/code_census or "
        'Set DB URL as environment variable like DB_URL="postgresql://postgres:pass@db.host:5432/code_census".'
    )
}

error_console = Console(stderr=True, style="bold red")
info_console = Console(style="black green")


def set_db_url(url):
    os.environ['DB_URL'] = url


class JSONType(click.ParamType):
    name = "json"

    def convert(self, value, param, ctx):
        if isinstance(value, (dict, list)):
            return value
        try:
            converted_value = json.loads(value)
        except json.decoder.JSONDecodeError:
            self.fail(f"{value=} is not a valid JSON"), param, ctx

    def __repr__(self) -> str:
        return "JSON"


@click.group()
def cli():
    pass


@cli.group()
def mypy():
    pass


@cli.group()
def project():
    pass


@project.command()
@click.argument("name", type=str, required=True)
@click.option("--description", type=str, default="")
@click.option("--url", type=str, default="")
@click.option(
    "--db-url", type=str, required=True, envvar="DB_URL", help=HELP_TEXT["db_url"]
)
def create(name: str, description: str, url: str, db_url: str):
    set_db_url(db_url)
    name = name.strip()
    session = create_session(db_url, echo=False)
    project = get_project(session, name=name)

    if project:
        error_console.print(f"Project {project.name} already exists")
        sys.exit(-1)

    project = Project(name=name, description=description, url=url)
    session.add(project)
    session.commit()
    info_console.print(f"Project created. {project=}")


@project.command()
@click.option(
    "--db-url", type=str, required=True, envvar="DB_URL", help=HELP_TEXT["db_url"]
)
def all(db_url):
    set_db_url(db_url)
    session = create_session(db_url, echo=False)
    projects = get_projects(session)
    table = formatter.format_projects(projects)
    info_console.print(table)


@mypy.group()
def run():
    pass


@run.command()
@click.argument("project_name", type=str, required=True)
@click.option("--artifact-url", type=str, required=False, default="")
@click.option("--run-info", type=JSONType(), default={})
@click.option("--mypy-coverage-file", type=click.Path(), required=True)
@click.option(
    "--db-url", type=str, required=True, envvar="DB_URL", help=HELP_TEXT["db_url"]
)
def add(
    project_name: str,
    artifact_url: str,
    run_info: dict,
    mypy_coverage_file: click.Path,
    db_url: str,
    log=True,
):
    set_db_url(db_url)
    name = project_name.strip()
    session = create_session(db_url, echo=log)
    project = get_project(session=session, name=project_name)

    if not project:
        error_console.print(f"{project_name=} is missing.")
        info_console.print("Create one using, [bold] project create name [/bold]")
        sys.exit(-1)

    run = add_run(
        session=session, project=project, artifact_url=artifact_url, run_info=run_info
    )

    cov_filename = Path(mypy_coverage_file)
    if cov_filename.exists():
        summaries = get_type_coverage(cov_filename=mypy_coverage_file)
        res = add_mypy_line_items(
            session=session, project=project, run=run, file_summaries=summaries
        )
        session.commit()
        count = len(res)
        info_console.print(f"Created a new {run=}")
        info_console.print(f"Added {count} file coverages")

    else:
        error_console.print(f"{mypy_coverage_file=} is missing")
        sys.exit(-1)


@run.command()
@click.option(
    "--db-url", type=str, required=True, envvar="DB_URL", help=HELP_TEXT["db_url"]
)
@click.argument("run_id", type=int)
def get_info(db_url: str, run_id: int):
    set_db_url(db_url)
    session = create_session(db_url, echo=False)
    items = get_mypy_line_items_by_run_id(session=session, run_id=run_id)

    if not items:
        error_console.print(f"[i] No run found for {run_id=} [/i]")
        sys.exit(-1)

    table = formatter.format_mypy_items(run_id, items)
    info_console.print(table)


@run.command()
@click.option(
    "--db-url", type=str, required=True, envvar="DB_URL", help=HELP_TEXT["db_url"]
)
@click.argument("project_name", type=str)
def all(db_url: str, project_name: str):
    set_db_url(db_url)
    session = create_session(db_url, echo=False)
    project_name = project_name.strip()
    project = get_project(session, name=project_name)

    if not project:
        # Print list of available project names
        error_console.print("{project_name=} is missing")
        sys.exit(-1)
    runs = get_runs(session=session, project=project)

    table = Table(title=f"All runs for project: {project.name}")
    table.add_column("ID")
    table.add_column("Created At")
    table.add_column("Run Info")
    table.add_column("Line Items")

    for run in runs:
        # N+1 query but it's fine for now
        count = len(run.mypylineitems)
        table.add_row(f"{run.id}", f"{run.created}", f"{run.run_info}", f"{count}")
    info_console.print(table)


@cli.command()
@click.option(
    "--db-url", type=str, required=True, envvar="DB_URL", help=HELP_TEXT["db_url"]
)
def create_db(db_url: str):
    echo = os.getenv("DB_ECHO", False)
    set_db_url(db_url)
    engine = create_engine(db_url, echo=echo)
    cfg = Config("alembic.ini")
    with engine.begin() as connection:
        cfg.attributes["connection"] = connection
        try:
            command.upgrade(cfg, "head")
            info_console.print(":white_check_mark: The schema for census app is created")
        except sqlalchemy.exc.ProgrammingError as exc:
            error_console.print(":x: Failed to create schema :x:")
            error_console.print(str(exc))


if __name__ == "__main__":
    cli()



"""Mypy coverage html file utils
"""
from dataclasses import dataclass
from pathlib import Path
from bs4 import BeautifulSoup  # type: ignore
import re


@dataclass
class FileSummary:
    full_filename: str
    imprecision: float
    lines: int


class ClassName:
    FILENAME = "summary-filename"
    IMPRECISION = "summary-precision"
    LINES = "summary-lines"


INT_REGEX = "[-+]?[0-9]*\.?[0-9]+(?:[eE][-+]?[0-9]+)?"


def get_type_coverage(cov_filename: Path) -> list[FileSummary]:
    with open(cov_filename) as fp:
        soup = BeautifulSoup(fp, "html.parser")

    run_summary: RunSummary = None  # type: ignore
    file_summaries: list[FileSummary] = []

    for tr in soup.table.find_all("tr"):
        # Get all the file summaries
        full_filename, imprecision, lines, top_level_module = "", 0.00, 0, ""
        for td in tr.find_all("td"):
            val = td.get_text()
            classes = td.get("class")
            if ClassName.FILENAME in classes:
                full_filename = val

            elif ClassName.IMPRECISION in classes:
                vals = re.findall(INT_REGEX, val)
                if vals:
                    imprecision = float(vals[0])

            elif ClassName.LINES in classes:
                vals = re.findall(INT_REGEX, val)
                if vals:
                    lines = vals[0]

        if full_filename:
            file_summaries.append(
                FileSummary(
                    full_filename=full_filename, imprecision=imprecision, lines=lines
                )
            )
    return file_summaries



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['code_census']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy-Utils>=0.37.8,<0.38.0',
 'SQLAlchemy==1.4.0',
 'alembic>=1.6.5,<2.0.0',
 'beautifulsoup4>=4.9.3,<5.0.0',
 'click',
 'lxml>=4.6.3,<5.0.0',
 'mypy>=0.910,<0.911',
 'psycopg2>=2.9.1,<3.0.0',
 'rich>=10.7.0,<11.0.0']

entry_points = \
{'console_scripts': ['census = code_census.cli:cli',
                     'code_census = code_census.cli:cli']}

setup_kwargs = {
    'name': 'code-census',
    'version': '0.0.1',
    'description': 'A command line tool to collect, organize, report code metrics.',
    'long_description': None,
    'author': 'Kracekumar',
    'author_email': 'me@kracekumar.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from sqlalchemy_utils import URLType, Timestamp
import sqlalchemy as sa
from sqlalchemy import create_engine
from typing import Optional
from sqlalchemy.orm import declarative_base, sessionmaker, relationship
from .mypy_utils import FileSummary

RUN_INFO_DESCRIPTION = """Store run details as a dictionary.

Example
-------
{
  "git": {"branch": "main", "commit_id": "8eea08c125b26aed539166207bda2f7ca1982493",
          "pr": 1},
  "execution_time": {"duration": 500, "unit": "sec"},
  "mypy": {"version": 0.912}
}
"""

Base = declarative_base()


class IDMixin(object):
    """Base class for all the models"""

    id = sa.Column(sa.BigInteger, primary_key=True)


class Project(Base, Timestamp, IDMixin):
    __tablename__ = "project"

    name = sa.Column(sa.String, unique=True, index=True)
    description = sa.Column(sa.String, nullable=True, default="")
    url = sa.Column(URLType, default="")

    def __repr__(self):
        return (
            f"<Project(id={self.id}, name={self.name}, description={self.description})>"
        )


class Run(Base, Timestamp, IDMixin):
    __tablename__ = "run"

    project_id = sa.Column(sa.Integer, sa.ForeignKey(Project.id))
    artifact_url = sa.Column(URLType, nullable=False, default="")
    run_info = sa.Column(sa.JSON)

    project = relationship(Project)
    mypylineitems = relationship("MypyRunLineItem")

    def __repr__(self):
        return f"<Run(id={self.id}, project_name={self.project.name})"


class MypyRunLineItem(Base, Timestamp, IDMixin):
    __tablename__ = "mypyrunlineitem"

    project_id = sa.Column(sa.Integer, sa.ForeignKey(Project.id))
    run_id = sa.Column(sa.Integer, sa.ForeignKey(Run.id))
    # "Path(foo/bar/custom.py) or file path representation(foo.bar.custom)"
    path = sa.Column(sa.Unicode)
    loc = sa.Column(sa.Integer)
    imprecision = sa.Column(sa.Float)

    project = relationship(Project)
    run = relationship(Run)

    def __repr__(self):
        return f"<MyprRunLineItem(path={self.path}, loc={self.loc}, imprecision={self.imprecision})"


def create_session(db_url: str, echo: bool = True):
    engine = create_engine(db_url, echo=echo)
    Session = sessionmaker(bind=engine)
    return Session()


def get_project(session, name: str) -> Optional[Project]:
    return session.query(Project).filter_by(name=name.strip()).first()


def get_projects(session) -> list[Project]:
    return session.query(Project).all()


def get_runs(session, project: Project) -> list[Run]:
    return session.query(Run).filter_by(project=project).order_by(Run.id.desc()).all()


def add_run(session, project: Project, artifact_url: str, run_info: dict) -> Run:
    run = Run(project=project, artifact_url=artifact_url, run_info=run_info)
    session.add(run)
    return run


def add_mypy_line_item(session, project: Project, run: Run, file_summary: FileSummary):
    item = MypyRunLineItem(
        project=project,
        run=run,
        path=file_summary.full_filename,
        loc=file_summary.lines,
        imprecision=file_summary.imprecision,
    )
    session.add(item)
    return item


def add_mypy_line_items(
    session, project: Project, run: Run, file_summaries: list[FileSummary]
) -> list[MypyRunLineItem]:
    summaries = []
    for summary in file_summaries:
        item = add_mypy_line_item(
            session=session, project=project, run=run, file_summary=summary
        )
        summaries.append(item)
    return summaries


def get_mypy_line_items_by_run_id(session, run_id: int) -> list[MypyRunLineItem]:
    return session.query(MypyRunLineItem).filter_by(run_id=run_id).all()
